@@ -1,4 +1,4 @@
-from typing import Any
+from typing import Any, Union
 
 from .exceptions import NotSupportedError
 
@@ -10,10 +10,13 @@ class KeyInterface:
     :func:`pyseto.Key.new <pyseto.Key.new>` returns an object which has this interface.
     """
 
-    def __init__(self, version: int, type: str, key: Any):
+    _VERSION = 0
+    _TYPE = ""
+
+    def __init__(self, version: int, purpose: str, key: Any):
         self._version = version
-        self._type = type
-        self._header = (f"v{self._version}." + self._type + ".").encode("utf-8")
+        self._purpose = purpose
+        self._header = (f"v{self._version}." + self._purpose + ".").encode("utf-8")
         self._sig_size = 0
         self._key: Any = key
         if not self._key:
@@ -28,20 +31,75 @@ class KeyInterface:
         return self._version
 
     @property
-    def type(self) -> str:
+    def purpose(self) -> str:
         """
-        The type (purpose) of the key. It will be ``"local"`` or ``"public"``.
+        The purpose of the key. It will be ``"local"`` or ``"public"``.
         """
-        return self._type
+        return self._purpose
 
     @property
     def header(self) -> bytes:
         """
-        The header value for a PASETO token. It will be ``"v<version>.<type>."``.
+        The header value for a PASETO token. It will be ``"v<version>.<purpose>."``.
         For example, ``"v1.local."``.
         """
         return self._header
 
+    # @property
+    # def key(self) -> Any:
+    #     """
+    #     Byte string or pyca/cryptography key object of the key.
+    #     If the key is ``public``, the key object is one of the pyca/cryptography objects bellow:
+
+    #     - ``RSAPrivateKey``
+    #     - ``RSAPublicKey``
+    #     - ``Ed25519PrivateKey``
+    #     - ``Ed25519PublicKey``
+    #     - ``EllipticCurvePrivateKey``
+    #     - ``EllipticCurvePublicKey``
+    #     """
+    #     return self._key
+
+    def to_paserk(
+        self,
+        wrapping_key: Union[bytes, str] = b"",
+        password: Union[bytes, str] = b"",
+        iteration: int = 100000,
+        memory_cost: int = 15 * 1024,
+        time_cost: int = 2,
+        parallelism: int = 1,
+    ) -> str:
+        """
+        Returns the PASERK expression of the key.
+
+        Args:
+            wrapping_key (Union[bytes, str]): A wrapping key to wrap the key.
+                If the `wrapping_key` is specified, `password` should not be
+                specified.
+            password (Union[bytes, str]): A password to wrap the key. If the
+                `password` is specified, `wrapping_key` should not be specified.
+            iteration (int): An iteration count used for password-based key
+                wrapping. This argument will only be used when the `password` is
+                specified.
+        Returns:
+            str: A PASERK string.
+        Raise:
+            ValueError: Invalid arguments.
+            EncryptError: Failed to wrap the key.
+        """
+        raise NotImplementedError(
+            "The PASERK expression for the key is not supported yet."
+        )
+
+    def to_paserk_id(self) -> str:
+        """
+        Returns the PASERK ID of the key.
+
+        Returns:
+            str: A PASERK ID string.
+        """
+        raise NotImplementedError("The PASERK ID for the key is not supported yet.")
+
     def encrypt(
         self,
         payload: bytes,
@@ -143,23 +201,3 @@ class KeyInterface:
             NotSupportedError: The key does not support the operation.
         """
         raise NotSupportedError("A key for local does not have verify().")
-
-    def to_paserk(self) -> str:
-        """
-        Returns the PASERK expression of the key.
-
-        Returns:
-            str: A PASERK string.
-        """
-        raise NotImplementedError(
-            "The PASERK expression for the key is not supported yet."
-        )
-
-    def to_paserk_id(self) -> str:
-        """
-        Returns the PASERK ID of the key.
-
-        Returns:
-            str: A PASERK ID string.
-        """
-        raise NotImplementedError("The PASERK ID for the key is not supported yet.")
