@@ -6,37 +6,38 @@ from typing import Any, Union
 from cryptography.hazmat.primitives import hashes, serialization
 from cryptography.hazmat.primitives.asymmetric import padding
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
-from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
 from cryptography.hazmat.primitives.kdf.hkdf import HKDF
 from cryptography.hazmat.primitives.serialization import (
     load_der_private_key,
     load_der_public_key,
 )
 
-from ..exceptions import DecryptError, EncryptError, SignError, VerifyError
+from ..exceptions import DecryptError, SignError, VerifyError
 from ..key_interface import KeyInterface
-from ..local_key import LocalKey
+from ..key_nist import NISTKey
 from ..utils import base64url_decode, base64url_encode, pae
 
 
-class V1Local(LocalKey):
+class V1Local(NISTKey):
     """
     The key object for v1.local.
     """
 
+    _VERSION = 1
+    _TYPE = "local"
+
     def __init__(self, key: Union[str, bytes]):
 
-        super().__init__(1, "local", key)
+        super().__init__(key)
         return
 
-    @classmethod
-    def from_paserk(cls, paserk: str) -> KeyInterface:
-        frags = paserk.split(".")
-        if frags[0] != "k1":
-            raise ValueError("Invalid PASERK version for a v1.local key.")
-        if frags[1] != "local":
-            raise ValueError("Invalid PASERK type for a v1.local key.")
-        return cls(base64url_decode(frags[2]))
+    def to_paserk_id(self) -> str:
+        h = "k1.lid."
+        p = self.to_paserk()
+        digest = hashes.Hash(hashes.SHA384())
+        digest.update((h + p).encode("utf-8"))
+        d = digest.finalize()
+        return h + base64url_encode(d[0:33]).decode("utf-8")
 
     def encrypt(
         self,
@@ -46,7 +47,13 @@ class V1Local(LocalKey):
         nonce: bytes = b"",
     ) -> bytes:
 
-        n = self._get_nonce(payload, nonce)
+        if nonce:
+            if len(nonce) != 32:
+                raise ValueError("nonce must be 32 bytes long.")
+        else:
+            nonce = token_bytes(32)
+
+        n = self._generate_hash(nonce, payload, 32)
         e = HKDF(
             algorithm=hashes.SHA384(),
             length=32,
@@ -62,20 +69,13 @@ class V1Local(LocalKey):
         ek = e.derive(self._key)
         ak = a.derive(self._key)
 
-        try:
-            c = (
-                Cipher(algorithms.AES(ek), modes.CTR(n[16:]))
-                .encryptor()
-                .update(payload)
-            )
-            pre_auth = pae([self.header, n, c, footer])
-            t = hmac.new(ak, pre_auth, hashlib.sha384).digest()
-            token = self._header + base64url_encode(n + c + t)
-            if footer:
-                token += b"." + base64url_encode(footer)
-            return token
-        except Exception as err:
-            raise EncryptError("Failed to encrypt.") from err
+        c = self._encrypt(ek, n[16:], payload)
+        pre_auth = pae([self.header, n, c, footer])
+        t = hmac.new(ak, pre_auth, hashlib.sha384).digest()
+        token = self._header + base64url_encode(n + c + t)
+        if footer:
+            token += b"." + base64url_encode(footer)
+        return token
 
     def decrypt(
         self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
@@ -103,44 +103,22 @@ class V1Local(LocalKey):
         t2 = hmac.new(ak, pre_auth, hashlib.sha384).digest()
         if t != t2:
             raise DecryptError("Failed to decrypt.")
-
-        try:
-            return Cipher(algorithms.AES(ek), modes.CTR(n[16:])).decryptor().update(c)
-        except Exception as err:
-            raise DecryptError("Failed to decrypt.") from err
-
-    def to_paserk_id(self) -> str:
-        h = "k1.lid."
-        p = self.to_paserk()
-        digest = hashes.Hash(hashes.SHA384())
-        digest.update((h + p).encode("utf-8"))
-        d = digest.finalize()
-        return h + base64url_encode(d[0:33]).decode("utf-8")
-
-    def _get_nonce(self, msg: bytes, nonce: bytes = b"") -> bytes:
-
-        if nonce:
-            if len(nonce) != 32:
-                raise ValueError("nonce must be 32 bytes long.")
-        else:
-            nonce = token_bytes(32)
-
-        try:
-            return hmac.new(nonce, msg, hashlib.sha384).digest()[0:32]
-        except Exception as err:
-            raise EncryptError("Failed to get nonce.") from err
+        return self._decrypt(ek, n[16:], c)
 
 
-class V1Public(KeyInterface):
+class V1Public(NISTKey):
     """
     The key object for v1.public.
     """
 
+    _VERSION = 1
+    _TYPE = "public"
+
     def __init__(self, key: Any):
 
-        super().__init__(1, "public", key)
-        self._sig_size = 256
+        super().__init__(key)
 
+        self._sig_size = 256
         if not isinstance(self._key, (RSAPublicKey, RSAPrivateKey)):
             raise ValueError("The key is not RSA key.")
 
@@ -148,15 +126,123 @@ class V1Public(KeyInterface):
         return
 
     @classmethod
-    def from_paserk(cls, paserk: str) -> KeyInterface:
+    def from_paserk(
+        cls, paserk: str, wrapping_key: bytes = b"", password: bytes = b""
+    ) -> KeyInterface:
+
+        if wrapping_key and password:
+            raise ValueError(
+                "Only one of wrapping_key or password should be specified."
+            )
+
         frags = paserk.split(".")
         if frags[0] != "k1":
-            raise ValueError("Invalid PASERK version for a v1.public key.")
+            raise ValueError(f"Invalid PASERK version: {frags[0]}.")
+
+        if wrapping_key:
+            if len(frags) != 4:
+                raise ValueError("Invalid PASERK format.")
+            if frags[2] != "pie":
+                raise ValueError(f"Unknown wrapping algorithm: {frags[2]}.")
+
+            if frags[1] == "secret-wrap":
+                h = "k1.secret-wrap.pie."
+                k = cls._decode_pie(h, wrapping_key, frags[3])
+                return cls(load_der_private_key(k, password=None))
+            raise ValueError(f"Invalid PASERK type: {frags[1]}.")
+
+        if len(frags) != 3:
+            raise ValueError("Invalid PASERK format.")
+
+        if password:
+            if len(frags) != 3:
+                raise ValueError("Invalid PASERK format.")
+
+            if frags[1] == "secret-pw":
+                h = "k1.secret-pw."
+                k = cls._decode_pbkw(h, password, frags[2])
+                return cls(load_der_private_key(k, password=None))
+            raise ValueError(f"Invalid PASERK type: {frags[1]}.")
+
+        wrapped = base64url_decode(frags[2])
         if frags[1] == "public":
-            return cls(load_der_public_key(base64url_decode(frags[2])))
-        elif frags[1] == "secret":
-            return cls(load_der_private_key(base64url_decode(frags[2]), password=None))
-        raise ValueError("Invalid PASERK type for a v1.public key.")
+            return cls(load_der_public_key(wrapped))
+        if frags[1] == "secret":
+            return cls(load_der_private_key(wrapped, password=None))
+        if frags[1] == "secret-wrap":
+            raise ValueError(f"{frags[1]} needs wrapping_key.")
+        raise ValueError(f"Invalid PASERK type: {frags[1]}.")
+
+    def to_paserk(
+        self,
+        wrapping_key: Union[bytes, str] = b"",
+        password: Union[bytes, str] = b"",
+        iteration: int = 100000,
+        memory_cost: int = 15 * 1024,
+        time_cost: int = 2,
+        parallelism: int = 1,
+    ) -> str:
+
+        if wrapping_key and password:
+            raise ValueError(
+                "Only one of wrapping_key or password should be specified."
+            )
+
+        if wrapping_key:
+            # secret-wrap
+            if not isinstance(self._key, RSAPrivateKey):
+                raise ValueError("Public key cannot be wrapped.")
+
+            bkey = (
+                wrapping_key
+                if isinstance(wrapping_key, bytes)
+                else wrapping_key.encode("utf-8")
+            )
+            h = "k1.secret-wrap.pie."
+            priv = self._key.private_bytes(
+                encoding=serialization.Encoding.DER,
+                format=serialization.PrivateFormat.TraditionalOpenSSL,
+                encryption_algorithm=serialization.NoEncryption(),
+            )
+            return h + self._encode_pie(h, bkey, priv)
+
+        if password:
+            # secret-pw
+            if not isinstance(self._key, RSAPrivateKey):
+                raise ValueError("Public key cannot be wrapped.")
+
+            bpw = password if isinstance(password, bytes) else password.encode("utf-8")
+            h = "k1.secret-pw."
+            priv = self._key.private_bytes(
+                encoding=serialization.Encoding.DER,
+                format=serialization.PrivateFormat.TraditionalOpenSSL,
+                encryption_algorithm=serialization.NoEncryption(),
+            )
+            return h + self._encode_pbkw(h, bpw, priv, iteration)
+
+        # public
+        if isinstance(self._key, RSAPublicKey):
+            pub = self._key.public_bytes(
+                encoding=serialization.Encoding.DER,
+                format=serialization.PublicFormat.SubjectPublicKeyInfo,
+            )
+            return "k1.public." + base64url_encode(pub).decode("utf-8")
+
+        # secret
+        priv = self._key.private_bytes(
+            encoding=serialization.Encoding.DER,
+            format=serialization.PrivateFormat.TraditionalOpenSSL,
+            encryption_algorithm=serialization.NoEncryption(),
+        )
+        return "k1.secret." + base64url_encode(priv).decode("utf-8")
+
+    def to_paserk_id(self) -> str:
+        p = self.to_paserk()
+        h = "k1.pid." if isinstance(self._key, RSAPublicKey) else "k1.sid."
+        digest = hashes.Hash(hashes.SHA384())
+        digest.update((h + p).encode("utf-8"))
+        d = digest.finalize()
+        return h + base64url_encode(d[0:33]).decode("utf-8")
 
     def sign(
         self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
@@ -186,34 +272,3 @@ class V1Public(KeyInterface):
         except Exception as err:
             raise VerifyError("Failed to verify.") from err
         return m
-
-    def to_paserk(self, seed: bytes = b"") -> str:
-        if isinstance(self._key, RSAPublicKey):
-            return (
-                "k1.public."
-                + base64url_encode(
-                    self._key.public_bytes(
-                        encoding=serialization.Encoding.DER,
-                        format=serialization.PublicFormat.SubjectPublicKeyInfo,
-                    )
-                ).decode("utf-8")
-            )
-        return (
-            "k1.secret."
-            + base64url_encode(
-                seed
-                + self._key.private_bytes(
-                    encoding=serialization.Encoding.DER,
-                    format=serialization.PrivateFormat.TraditionalOpenSSL,
-                    encryption_algorithm=serialization.NoEncryption(),
-                )
-            ).decode("utf-8")
-        )
-
-    def to_paserk_id(self) -> str:
-        p = self.to_paserk()
-        h = "k1.pid." if isinstance(self._key, RSAPublicKey) else "k1.sid."
-        digest = hashes.Hash(hashes.SHA384())
-        digest.update((h + p).encode("utf-8"))
-        d = digest.finalize()
-        return h + base64url_encode(d[0:33]).decode("utf-8")
