@@ -3,7 +3,7 @@ import hmac
 from secrets import token_bytes
 from typing import Any, Union
 
-from cryptography.hazmat.primitives import hashes, serialization
+from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.primitives.asymmetric import ec
 from cryptography.hazmat.primitives.asymmetric.ec import (
     EllipticCurvePrivateKey,
@@ -13,34 +13,27 @@ from cryptography.hazmat.primitives.asymmetric.utils import (
     decode_dss_signature,
     encode_dss_signature,
 )
-from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
 from cryptography.hazmat.primitives.kdf.hkdf import HKDF
 
-from ..exceptions import DecryptError, EncryptError, SignError, VerifyError
+from ..exceptions import DecryptError, SignError, VerifyError
 from ..key_interface import KeyInterface
-from ..local_key import LocalKey
+from ..key_nist import NISTKey
 from ..utils import base64url_decode, base64url_encode, i2osp, os2ip, pae
 
 
-class V3Local(LocalKey):
+class V3Local(NISTKey):
     """
     The key object for v3.local.
     """
 
+    _VERSION = 3
+    _TYPE = "local"
+
     def __init__(self, key: Union[str, bytes]):
 
-        super().__init__(3, "local", key)
+        super().__init__(key)
         return
 
-    @classmethod
-    def from_paserk(cls, paserk: str) -> KeyInterface:
-        frags = paserk.split(".")
-        if frags[0] != "k3":
-            raise ValueError("Invalid PASERK version for a v3.local key.")
-        if frags[1] != "local":
-            raise ValueError("Invalid PASERK type for a v3.local key.")
-        return cls(base64url_decode(frags[2]))
-
     def encrypt(
         self,
         payload: bytes,
@@ -74,16 +67,13 @@ class V3Local(LocalKey):
         except Exception as err:
             raise DecryptError("Failed to derive keys.") from err
 
-        try:
-            c = Cipher(algorithms.AES(ek), modes.CTR(n2)).encryptor().update(payload)
-            pre_auth = pae([self.header, nonce, c, footer, implicit_assertion])
-            t = hmac.new(ak, pre_auth, hashlib.sha384).digest()
-            token = self._header + base64url_encode(nonce + c + t)
-            if footer:
-                token += b"." + base64url_encode(footer)
-            return token
-        except Exception as err:
-            raise EncryptError("Failed to encrypt.") from err
+        c = self._encrypt(ek, n2, payload)
+        pre_auth = pae([self.header, nonce, c, footer, implicit_assertion])
+        t = hmac.new(ak, pre_auth, hashlib.sha384).digest()
+        token = self._header + base64url_encode(nonce + c + t)
+        if footer:
+            token += b"." + base64url_encode(footer)
+        return token
 
     def decrypt(
         self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
@@ -116,13 +106,10 @@ class V3Local(LocalKey):
         t2 = hmac.new(ak, pre_auth, hashlib.sha384).digest()
         if t != t2:
             raise DecryptError("Failed to decrypt.")
-
-        try:
-            return Cipher(algorithms.AES(ek), modes.CTR(n2)).decryptor().update(c)
-        except Exception as err:
-            raise DecryptError("Failed to decrypt a message.") from err
+        return self._decrypt(ek, n2, c)
 
     def to_paserk_id(self) -> str:
+
         h = "k3.lid."
         p = self.to_paserk()
         digest = hashes.Hash(hashes.SHA384())
@@ -131,35 +118,23 @@ class V3Local(LocalKey):
         return h + base64url_encode(d[0:33]).decode("utf-8")
 
 
-class V3Public(KeyInterface):
+class V3Public(NISTKey):
     """
     The key object for v3.public.
     """
 
+    _VERSION = 3
+    _TYPE = "public"
+
     def __init__(self, key: Any):
 
-        super().__init__(3, "public", key)
-        self._sig_size = 96
+        super().__init__(key)
 
+        self._sig_size = 96
         if not isinstance(self._key, (EllipticCurvePublicKey, EllipticCurvePrivateKey)):
             raise ValueError("The key is not ECDSA key.")
         return
 
-    @classmethod
-    def from_paserk(cls, paserk: str) -> KeyInterface:
-        frags = paserk.split(".")
-        if frags[0] != "k3":
-            raise ValueError("Invalid PASERK version for a v3.public key.")
-        if frags[1] == "public":
-            return cls(
-                EllipticCurvePublicKey.from_encoded_point(
-                    ec.SECP384R1(), base64url_decode(frags[2])
-                )
-            )
-        # elif frags[1] == "secret":
-        #     return cls(Ed25519PrivateKey.from_private_bytes(base64url_decode(frags[2])))
-        raise ValueError("Invalid PASERK type for a v3.public key.")
-
     @classmethod
     def from_public_bytes(cls, key: bytes):
         try:
@@ -168,6 +143,124 @@ class V3Public(KeyInterface):
             raise ValueError("Invalid bytes for the key.") from err
         return cls(k)
 
+    @classmethod
+    def from_paserk(
+        cls, paserk: str, wrapping_key: bytes = b"", password: bytes = b""
+    ) -> KeyInterface:
+
+        if wrapping_key and password:
+            raise ValueError(
+                "Only one of wrapping_key or password should be specified."
+            )
+
+        frags = paserk.split(".")
+        if frags[0] != "k3":
+            raise ValueError(f"Invalid PASERK version: {frags[0]}.")
+
+        if wrapping_key:
+            # secret-wrap
+            if len(frags) != 4:
+                raise ValueError("Invalid PASERK format.")
+            if frags[2] != "pie":
+                raise ValueError(f"Unknown wrapping algorithm: {frags[2]}.")
+
+            if frags[1] == "secret-wrap":
+                h = "k3.secret-wrap.pie."
+                k = cls._decode_pie(h, wrapping_key, frags[3])
+                priv = ec.derive_private_key(
+                    int.from_bytes(k, byteorder="big"), ec.SECP384R1()
+                )
+                return cls(priv)
+            raise ValueError(f"Invalid PASERK type: {frags[1]}.")
+
+        if len(frags) != 3:
+            raise ValueError("Invalid PASERK format.")
+
+        if password:
+            # secret-pw
+            if frags[1] == "secret-pw":
+                h = "k3.secret-pw."
+                k = cls._decode_pbkw(h, password, frags[2])
+                priv = ec.derive_private_key(
+                    int.from_bytes(k, byteorder="big"), ec.SECP384R1()
+                )
+                return cls(priv)
+            raise ValueError(f"Invalid PASERK type: {frags[1]}.")
+
+        # public
+        k = base64url_decode(frags[2])
+        if frags[1] == "public":
+            pub = EllipticCurvePublicKey.from_encoded_point(ec.SECP384R1(), k)
+            return cls(pub)
+
+        # secret
+        if frags[1] == "secret":
+            priv = ec.derive_private_key(
+                int.from_bytes(k, byteorder="big"), ec.SECP384R1()
+            )
+            return cls(priv)
+        if frags[1] == "secret-wrap":
+            raise ValueError(f"{frags[1]} needs wrapping_key.")
+        raise ValueError(f"Invalid PASERK type: {frags[1]}.")
+
+    def to_paserk(
+        self,
+        wrapping_key: Union[bytes, str] = b"",
+        password: Union[bytes, str] = b"",
+        iteration: int = 100000,
+        memory_cost: int = 15 * 1024,
+        time_cost: int = 2,
+        parallelism: int = 1,
+    ) -> str:
+
+        if wrapping_key and password:
+            raise ValueError(
+                "Only one of wrapping_key or password should be specified."
+            )
+
+        if wrapping_key:
+            # secret-wrap
+            if not isinstance(self._key, EllipticCurvePrivateKey):
+                raise ValueError("Public key cannot be wrapped.")
+
+            bkey = (
+                wrapping_key
+                if isinstance(wrapping_key, bytes)
+                else wrapping_key.encode("utf-8")
+            )
+            h = "k3.secret-wrap.pie."
+            k = self._key.private_numbers().private_value.to_bytes(48, byteorder="big")
+            return h + self._encode_pie(h, bkey, k)
+
+        if password:
+            # secret-pw
+            if not isinstance(self._key, EllipticCurvePrivateKey):
+                raise ValueError("Public key cannot be wrapped.")
+
+            bpw = password if isinstance(password, bytes) else password.encode("utf-8")
+            h = "k3.secret-pw."
+            k = self._key.private_numbers().private_value.to_bytes(48, byteorder="big")
+            return h + self._encode_pbkw(h, bpw, k, iteration)
+
+        # public
+        if isinstance(self._key, EllipticCurvePublicKey):
+            k = self._public_key_compress(
+                self._key.public_numbers().x, self._key.public_numbers().y
+            )
+            return "k3.public." + base64url_encode(k).decode("utf-8")
+
+        # private
+        k = self._key.private_numbers().private_value.to_bytes(48, byteorder="big")
+        return "k3.secret." + base64url_encode(k).decode("utf-8")
+
+    def to_paserk_id(self) -> str:
+        p = self.to_paserk()
+        h = "k3.pid." if isinstance(self._key, EllipticCurvePublicKey) else "k3.sid."
+        digest = hashes.Hash(hashes.SHA384())
+        digest.update((h + p).encode("utf-8"))
+        d = digest.finalize()
+        return h + base64url_encode(d[0:33]).decode("utf-8")
+
     def sign(
         self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
     ) -> bytes:
@@ -208,31 +301,6 @@ class V3Public(KeyInterface):
             raise VerifyError("Failed to verify.") from err
         return m
 
-    def to_paserk(self) -> str:
-        if isinstance(self._key, EllipticCurvePublicKey):
-            data = self._public_key_compress(
-                self._key.public_numbers().x, self._key.public_numbers().y
-            )
-            return "k3.public." + base64url_encode(data).decode("utf-8")
-        return (
-            "k3.secret."
-            + base64url_encode(
-                self._key.private_bytes(
-                    encoding=serialization.Encoding.DER,
-                    format=serialization.PrivateFormat.TraditionalOpenSSL,
-                    encryption_algorithm=serialization.NoEncryption(),
-                )
-            ).decode("utf-8")
-        )
-
-    def to_paserk_id(self) -> str:
-        p = self.to_paserk()
-        h = "k3.pid." if isinstance(self._key, EllipticCurvePublicKey) else "k3.sid."
-        digest = hashes.Hash(hashes.SHA384())
-        digest.update((h + p).encode("utf-8"))
-        d = digest.finalize()
-        return h + base64url_encode(d[0:33]).decode("utf-8")
-
     def _public_key_compress(self, x: int, y: int) -> bytes:
 
         bx = x.to_bytes(48, byteorder="big")
