@@ -2,7 +2,7 @@ from secrets import token_bytes
 
 import pytest
 
-from pyseto import Key, NotSupportedError
+from pyseto import DecryptError, Key, NotSupportedError
 from pyseto.key_interface import KeyInterface
 from pyseto.utils import base64url_decode
 
@@ -15,7 +15,7 @@ class TestKey:
     """
 
     @pytest.mark.parametrize(
-        "version, type, key",
+        "version, purpose, key",
         [
             (1, "local", token_bytes(32)),
             (2, "local", token_bytes(32)),
@@ -23,11 +23,11 @@ class TestKey:
             (4, "local", token_bytes(32)),
         ],
     )
-    def test_key_new_local(self, version, type, key):
-        k = Key.new(version, type, key)
+    def test_key_new_local(self, version, purpose, key):
+        k = Key.new(version, purpose, key)
         assert isinstance(k, KeyInterface)
         assert k.version == version
-        assert k.type == type
+        assert k.purpose == purpose
         with pytest.raises(NotSupportedError) as err:
             k.sign(b"Hello world!")
             pytest.fail("Key.sign() should fail.")
@@ -38,7 +38,7 @@ class TestKey:
         assert "A key for local does not have verify()." in str(err.value)
 
     @pytest.mark.parametrize(
-        "version, type, key",
+        "version, purpose, key",
         [
             (1, "public", load_key("keys/private_key_rsa.pem")),
             (1, "public", load_key("keys/public_key_rsa.pem")),
@@ -50,11 +50,11 @@ class TestKey:
             (4, "public", load_key("keys/public_key_ed25519.pem")),
         ],
     )
-    def test_key_new_public(self, version, type, key):
-        k = Key.new(version, type, key)
+    def test_key_new_public(self, version, purpose, key):
+        k = Key.new(version, purpose, key)
         assert isinstance(k, KeyInterface)
         assert k.version == version
-        assert k.type == type
+        assert k.purpose == purpose
         with pytest.raises(NotSupportedError) as err:
             k.encrypt(b"Hello world!")
             pytest.fail("Key.sign() should fail.")
@@ -67,51 +67,51 @@ class TestKey:
     @pytest.mark.parametrize(
         "version, key, msg",
         [
-            ("v1", load_key("keys/private_key_ed25519.pem"), "The key is not RSA key."),
-            ("v1", load_key("keys/public_key_ed25519.pem"), "The key is not RSA key."),
+            (1, load_key("keys/private_key_ed25519.pem"), "The key is not RSA key."),
+            (1, load_key("keys/public_key_ed25519.pem"), "The key is not RSA key."),
             (
-                "v1",
+                1,
                 load_key("keys/private_key_ecdsa_p384.pem"),
                 "The key is not RSA key.",
             ),
             (
-                "v1",
+                1,
                 load_key("keys/public_key_ecdsa_p384.pem"),
                 "The key is not RSA key.",
             ),
-            ("v2", load_key("keys/private_key_rsa.pem"), "The key is not Ed25519 key."),
-            ("v2", load_key("keys/public_key_rsa.pem"), "The key is not Ed25519 key."),
+            (2, load_key("keys/private_key_rsa.pem"), "The key is not Ed25519 key."),
+            (2, load_key("keys/public_key_rsa.pem"), "The key is not Ed25519 key."),
             (
-                "v2",
+                2,
                 load_key("keys/private_key_ecdsa_p384.pem"),
                 "The key is not Ed25519 key.",
             ),
             (
-                "v2",
+                2,
                 load_key("keys/public_key_ecdsa_p384.pem"),
                 "The key is not Ed25519 key.",
             ),
             (
-                "v3",
+                3,
                 load_key("keys/private_key_ed25519.pem"),
                 "The key is not ECDSA key.",
             ),
             (
-                "v3",
+                3,
                 load_key("keys/public_key_ed25519.pem"),
                 "The key is not ECDSA key.",
             ),
-            ("v3", load_key("keys/private_key_rsa.pem"), "The key is not ECDSA key."),
-            ("v3", load_key("keys/public_key_rsa.pem"), "The key is not ECDSA key."),
-            ("v4", load_key("keys/private_key_rsa.pem"), "The key is not Ed25519 key."),
-            ("v4", load_key("keys/public_key_rsa.pem"), "The key is not Ed25519 key."),
+            (3, load_key("keys/private_key_rsa.pem"), "The key is not ECDSA key."),
+            (3, load_key("keys/public_key_rsa.pem"), "The key is not ECDSA key."),
+            (4, load_key("keys/private_key_rsa.pem"), "The key is not Ed25519 key."),
+            (4, load_key("keys/public_key_rsa.pem"), "The key is not Ed25519 key."),
             (
-                "v4",
+                4,
                 load_key("keys/private_key_ecdsa_p384.pem"),
                 "The key is not Ed25519 key.",
             ),
             (
-                "v4",
+                4,
                 load_key("keys/public_key_ecdsa_p384.pem"),
                 "The key is not Ed25519 key.",
             ),
@@ -124,7 +124,7 @@ class TestKey:
         assert msg in str(err.value)
 
     @pytest.mark.parametrize(
-        "version, type, key, msg",
+        "version, purpose, key, msg",
         [
             ("v*", "local", token_bytes(32), "Invalid version: v*."),
             ("v0", "local", token_bytes(32), "Invalid version: v0."),
@@ -147,13 +147,13 @@ class TestKey:
                 load_key("keys/private_key_rsa.pem"),
                 "Invalid version: 0.",
             ),
-            ("v1", "xxx", token_bytes(32), "Invalid type(purpose): xxx."),
-            ("v1", "public", "-----BEGIN BAD", "Invalid or unsupported PEM format."),
+            (1, "xxx", token_bytes(32), "Invalid purpose: xxx."),
+            (1, "public", "-----BEGIN BAD", "Invalid or unsupported PEM format."),
         ],
     )
-    def test_key_new_with_invalid_arg(self, version, type, key, msg):
+    def test_key_new_with_invalid_arg(self, version, purpose, key, msg):
         with pytest.raises(ValueError) as err:
-            Key.new(version, type, key)
+            Key.new(version, purpose, key)
             pytest.fail("Key.new should fail.")
         assert msg in str(err.value)
 
@@ -175,7 +175,49 @@ class TestKey:
         k = Key.from_asymmetric_key_params(version, x=key["x"], y=key["y"], d=key["d"])
         assert isinstance(k, KeyInterface)
         assert k.version == version
-        assert k.type == "public"
+        assert k.purpose == "public"
+
+    @pytest.mark.parametrize(
+        "paserk",
+        [
+            "k1.local.AAAAAAAAAAAAAAAA",
+            "k1.public.AAAAAAAAAAAAAAAA",
+            "k2.local.AAAAAAAAAAAAAAAA",
+            "k2.public.AAAAAAAAAAAAAAAA",
+            "k3.local.AAAAAAAAAAAAAAAA",
+            "k3.public.AAAAAAAAAAAAAAAA",
+            "k4.local.AAAAAAAAAAAAAAAA",
+            "k4.public.AAAAAAAAAAAAAAAA",
+        ],
+    )
+    def test_key_from_paserk_with_wrapping_key_and_password(self, paserk):
+
+        with pytest.raises(ValueError) as err:
+            Key.from_paserk(paserk, wrapping_key="xxx", password="yyy")
+            pytest.fail("Key.from_paserk should fail.")
+        assert "Only one of wrapping_key or password should be specified." in str(
+            err.value
+        )
+
+    @pytest.mark.parametrize(
+        "paserk, msg",
+        [
+            ("k1.local.AAAAAAAAAAAAAAAA", "Invalid PASERK type: local."),
+            ("k1.public.AAAAAAAAAAAAAAAA", "Invalid PASERK type: public."),
+            ("k2.local.AAAAAAAAAAAAAAAA", "Invalid PASERK type: local."),
+            ("k2.public.AAAAAAAAAAAAAAAA", "Invalid PASERK type: public."),
+            ("k3.local.AAAAAAAAAAAAAAAA", "Invalid PASERK type: local."),
+            ("k3.public.AAAAAAAAAAAAAAAA", "Invalid PASERK type: public."),
+            ("k4.local.AAAAAAAAAAAAAAAA", "Invalid PASERK type: local."),
+            ("k4.public.AAAAAAAAAAAAAAAA", "Invalid PASERK type: public."),
+        ],
+    )
+    def test_key_from_paserk_with_password_for_wrong_paserk(self, paserk, msg):
+
+        with pytest.raises(ValueError) as err:
+            Key.from_paserk(paserk, password="yyy")
+            pytest.fail("Key.from_paserk should fail.")
+        assert msg in str(err.value)
 
     @pytest.mark.parametrize(
         "paserk, msg",
@@ -192,41 +234,134 @@ class TestKey:
             pytest.fail("Key.from_paserk should fail.")
         assert msg in str(err.value)
 
+    @pytest.mark.parametrize(
+        "version",
+        [
+            1,
+            2,
+            3,
+            4,
+        ],
+    )
+    def test_key_from_paserk_for_local_with_wrong_wrapping_key(self, version):
+        k = Key.new(version, "local", token_bytes(32))
+        wk1 = token_bytes(32)
+        wk2 = token_bytes(32)
+        wpk = k.to_paserk(wrapping_key=wk1)
+        with pytest.raises(DecryptError) as err:
+            Key.from_paserk(wpk, wrapping_key=wk2)
+            pytest.fail("Key.from_paserk() should fail.")
+        assert "Failed to unwrap a key." in str(err.value)
+
+    @pytest.mark.parametrize(
+        "version",
+        [
+            1,
+            2,
+            3,
+            4,
+        ],
+    )
+    def test_key_from_paserk_for_local_with_wrong_password(self, version):
+        k = Key.new(version, "local", token_bytes(32))
+        wk1 = token_bytes(32)
+        wk2 = token_bytes(32)
+        wpk = k.to_paserk(password=wk1)
+        with pytest.raises(DecryptError) as err:
+            Key.from_paserk(wpk, password=wk2)
+            pytest.fail("Key.from_paserk() should fail.")
+        assert "Failed to unwrap a key." in str(err.value)
+
+    @pytest.mark.parametrize(
+        "version, key",
+        [
+            (1, load_key("keys/private_key_rsa.pem")),
+            (2, load_key("keys/private_key_ed25519.pem")),
+            (3, load_key("keys/private_key_ecdsa_p384.pem")),
+            (4, load_key("keys/private_key_ed25519.pem")),
+        ],
+    )
+    def test_key_from_paserk_for_private_key_with_wrong_wrapping_key(
+        self, version, key
+    ):
+        k = Key.new(version, "public", key)
+        wk1 = token_bytes(32)
+        wk2 = token_bytes(32)
+        wpk = k.to_paserk(wrapping_key=wk1)
+        with pytest.raises(DecryptError) as err:
+            Key.from_paserk(wpk, wrapping_key=wk2)
+            pytest.fail("Key.from_paserk() should fail.")
+        assert "Failed to unwrap a key." in str(err.value)
+
+    @pytest.mark.parametrize(
+        "version, key",
+        [
+            (1, load_key("keys/public_key_rsa.pem")),
+            (2, load_key("keys/public_key_ed25519.pem")),
+            (3, load_key("keys/public_key_ecdsa_p384.pem")),
+            (4, load_key("keys/public_key_ed25519.pem")),
+        ],
+    )
+    def test_key_from_paserk_for_public_key_with_wrapping_key(self, version, key):
+        k = Key.new(version, "public", key)
+        wk = token_bytes(32)
+        with pytest.raises(ValueError) as err:
+            k.to_paserk(wrapping_key=wk)
+            pytest.fail("to_paserk() should fail.")
+        assert "Public key cannot be wrapped." in str(err.value)
+
+    @pytest.mark.parametrize(
+        "version, key",
+        [
+            (1, load_key("keys/public_key_rsa.pem")),
+            (2, load_key("keys/public_key_ed25519.pem")),
+            (3, load_key("keys/public_key_ecdsa_p384.pem")),
+            (4, load_key("keys/public_key_ed25519.pem")),
+        ],
+    )
+    def test_key_from_paserk_for_public_key_with_password(self, version, key):
+        k = Key.new(version, "public", key)
+        wk = token_bytes(32)
+        with pytest.raises(ValueError) as err:
+            k.to_paserk(password=wk)
+            pytest.fail("to_paserk() should fail.")
+        assert "Public key cannot be wrapped." in str(err.value)
+
     @pytest.mark.parametrize(
         "version, key, msg",
         [
             (
-                "v1",
+                1,
                 load_jwk("keys/private_key_rsa.json"),
                 "v1.public is not supported on from_key_parameters.",
             ),
-            ("v*", load_jwk("keys/private_key_ed25519.json"), "Invalid version: v*."),
+            (999, load_jwk("keys/private_key_ed25519.json"), "Invalid version: 999."),
             (0, load_jwk("keys/private_key_ed25519.json"), "Invalid version: 0."),
             (
-                "v2",
+                2,
                 {"x": b"xxx", "y": b"", "d": b"ddd"},
                 "Only one of x or d should be set for v2.public.",
             ),
-            ("v2", {"x": b"xxx", "y": b"", "d": b""}, "Failed to load key."),
-            ("v2", {"x": b"", "y": b"", "d": b"ddd"}, "Failed to load key."),
+            (2, {"x": b"xxx", "y": b"", "d": b""}, "Failed to load key."),
+            (2, {"x": b"", "y": b"", "d": b"ddd"}, "Failed to load key."),
             (
-                "v2",
+                2,
                 {"x": b"", "y": b"", "d": b""},
                 "x or d should be set for v2.public.",
             ),
             (
-                "v3",
+                3,
                 {"x": b"xxx", "y": b"", "d": b"ddd"},
                 "x and y (and d) should be set for v3.public.",
             ),
             (
-                "v3",
+                3,
                 {"x": b"", "y": b"yyy", "d": b"ddd"},
                 "x and y (and d) should be set for v3.public.",
             ),
-            ("v3", {"x": b"xxx", "y": b"yyy", "d": b""}, "Failed to load key."),
+            (3, {"x": b"xxx", "y": b"yyy", "d": b""}, "Failed to load key."),
             (
-                "v3",
+                3,
                 {
                     "x": base64url_decode(
                         "_XyN9woHaS0mPimSW-etwJMEDSzxIMjp4PjezavU8SHJoClz1bQrcmPb1ZJxHxhI"
@@ -239,14 +374,14 @@ class TestKey:
                 "Failed to load key.",
             ),
             (
-                "v4",
+                4,
                 {"x": b"xxx", "y": b"", "d": b"ddd"},
                 "Only one of x or d should be set for v4.public.",
             ),
-            ("v4", {"x": b"xxx", "y": b"", "d": b""}, "Failed to load key."),
-            ("v4", {"x": b"", "y": b"", "d": b"ddd"}, "Failed to load key."),
+            (4, {"x": b"xxx", "y": b"", "d": b""}, "Failed to load key."),
+            (4, {"x": b"", "y": b"", "d": b"ddd"}, "Failed to load key."),
             (
-                "v4",
+                4,
                 {"x": b"", "y": b"", "d": b""},
                 "x or d should be set for v4.public.",
             ),
@@ -260,7 +395,7 @@ class TestKey:
         assert msg in str(err.value)
 
     @pytest.mark.parametrize(
-        "version, type, key",
+        "version, purpose, key",
         [
             (1, "public", load_key("keys/public_key_rsa.pem")),
             (2, "public", load_key("keys/public_key_ed25519.pem")),
@@ -268,12 +403,12 @@ class TestKey:
             (4, "public", load_key("keys/public_key_ed25519.pem")),
         ],
     )
-    def test_key_to_paserk_public(self, version, type, key):
-        k = Key.new(version, type, key)
+    def test_key_to_paserk_public(self, version, purpose, key):
+        k = Key.new(version, purpose, key)
         assert k.to_paserk().startswith(f"k{k.version}.public.")
 
     @pytest.mark.parametrize(
-        "version, type, key",
+        "version, purpose, key",
         [
             (1, "public", load_key("keys/private_key_rsa.pem")),
             (2, "public", load_key("keys/private_key_ed25519.pem")),
@@ -281,6 +416,30 @@ class TestKey:
             (4, "public", load_key("keys/private_key_ed25519.pem")),
         ],
     )
-    def test_key_to_paserk_secret(self, version, type, key):
-        k = Key.new(version, type, key)
+    def test_key_to_paserk_secret(self, version, purpose, key):
+        k = Key.new(version, purpose, key)
         assert k.to_paserk().startswith(f"k{k.version}.secret.")
+
+    @pytest.mark.parametrize(
+        "version, purpose, key",
+        [
+            (1, "local", token_bytes(32)),
+            (2, "local", token_bytes(32)),
+            (3, "local", token_bytes(32)),
+            (4, "local", token_bytes(32)),
+            (1, "public", load_key("keys/private_key_rsa.pem")),
+            (2, "public", load_key("keys/private_key_ed25519.pem")),
+            (3, "public", load_key("keys/private_key_ecdsa_p384.pem")),
+            (4, "public", load_key("keys/private_key_ed25519.pem")),
+        ],
+    )
+    def test_key_to_paserk_secret_with_wrapping_key_and_password(
+        self, version, purpose, key
+    ):
+        k = Key.new(version, purpose, key)
+        with pytest.raises(ValueError) as err:
+            k.to_paserk(wrapping_key="xxx", password="yyy")
+            pytest.fail("to_paserk() should fail.")
+        assert "Only one of wrapping_key or password should be specified." in str(
+            err.value
+        )
